/*
   get_proxy.c -- Get and check proxies

   Copyright (C) 2008 lorenzzz <lorenzzz@users.sourceforge.net>

   This file is part of Assh.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   $Id: get_proxy.c,v 1.16 2008/09/13 15:56:18 jlzz Exp $
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include "config.h"

char *text_to_parse=NULL;

extern char *cserver_hostname;
extern char *cserver_port;

void get_file_by_url(char *file) {
  strcpy(file,strchr(file+7,'/'));
}


void get_domain_by_url(char *dom) {
  strcpy(dom,strdup(dom)+7);
  *(strchr(dom,'/')) = '\0';
}


jmp_buf timeout_jump ;

void timeout(int sig) {
  longjmp(timeout_jump,1) ;
}


/* check_ssl(): If proxy allow ssl relay on dest port 22, return 0 else return 1 */
int  check_ssl(struct proxy *p){
  char buffer[100];        /* Command to send */
  char ans[200];            /* Proy answer */
  int sock;
  struct sockaddr_in dest;
  struct hostent *he;

  sprintf(buffer,"CONNECT %s:%s HTTP/1.0\r\n\n",cserver_hostname,cserver_port);

  if((he = gethostbyname(p->ip)) == NULL) {
    printf("Couldn't resolve %s!\n", p->ip);
    exit(-1);
  }

  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("socket()");
    exit(-1);
  }

  dest.sin_family = AF_INET;
  dest.sin_port = htons(atoi(p->port));
  dest.sin_addr = *((struct in_addr *)he->h_addr);

  signal(SIGALRM, timeout);
  alarm(CHECK_TIMEOUT);

  if(setjmp(timeout_jump)==1) {
    close(sock);
    return 1;
  } else {
    if(connect(sock, (struct sockaddr *)&dest,
               sizeof(struct sockaddr)) < 0) {
      alarm(0);
      close(sock);
      return 1;
    }
  }

  send(sock, buffer, strlen(buffer), 0);
  read(sock,ans,sizeof(ans));
  alarm(0);
  close(sock);

  if(strstr(ans,"200 Connection established")!=NULL)
    return 0;
  else
    return 1;
}


int get_proxy_list()
{
  struct hostent *he;
  struct sockaddr_in dest;
  int sock;

  int nread;
  char cmd[100];
  char *file,*domain;
  int size=0;

  file=malloc(strlen(PROXY_URL)+1);
  strncpy(file,PROXY_URL,strlen(PROXY_URL)+1);
  domain=strdup(file);

  get_file_by_url(file);
  get_domain_by_url(domain);

  if((he=gethostbyname(domain)) == NULL) {
    printf("[-] Cant resolve hostname\n");
    exit(1);
  }

  if((sock=socket(AF_INET,SOCK_STREAM,0)) == -1) {
    printf("[-] Cant create socket\n");
    exit(1);
  }

  dest.sin_family = AF_INET;
  dest.sin_port = htons(80);
  dest.sin_addr = *((struct in_addr *)he->h_addr);
  if(connect(sock, (struct sockaddr *)&dest, sizeof(struct sockaddr)) < 0) {
    printf("[-] connect() problem\n");
    exit(1); 
  }

  memset(cmd,0,sizeof(cmd));
  sprintf(cmd,"GET %s HTTP/1.0\nHost: %s\r\n\n",file,domain);
  if(write(sock,cmd,strlen(cmd)) == -1) {
    printf("[-] write() failed\n");
    exit(1);
  }

  sleep(2);

  memset(cmd,0,sizeof(cmd));
  while((nread=read(sock,cmd,sizeof(cmd)-1))>0) {
    size+=nread;
    text_to_parse=realloc(text_to_parse,size);
    strcat(text_to_parse,cmd);

    memset(cmd,0,sizeof(cmd));
  }
  close(sock);
  return 0;
}

int gen_proxy_list_from_file(char *srcfile,char *fileout){
  struct proxy *p;
  char buf[100],*ptr;
  FILE *fout, *fin;
  int pid;

  if((fin=fopen(srcfile,"r"))==NULL) {
    printf("[-] Can't open file %s\n",srcfile);
    exit(1);
  }

  p=malloc(sizeof(struct proxy));
  fout=fopen(fileout,"w+");

  while(fgets(buf,sizeof(buf),fin)!=NULL) {
    *(strchr(buf,'\n'))=0;
    ptr=strchr(buf,':');
    *ptr=0;
    strcpy(p->port,ptr+1);
    strcpy(p->ip,buf);
    if((pid=fork())==-1) {
      printf("[-] fork error\n");
      exit(1);
    } else if (pid==0) {
      if (check_ssl(p)==0)
        fprintf(fout,"%s:%s\n", p->ip, p->port);
      exit (0);
    }
  }
 
  free(p); 
  sleep(CHECK_TIMEOUT);
  fclose(fout);
  fclose(fin);

  return (EXIT_SUCCESS); 
}

int write_proxy_list (char *fileout)
{
  int err_addr, err_port;
  regex_t preg_addr, preg_port;
  const char *str_regex_addr = "([[:digit:]]{1,3}\\.[[:digit:]]{1,3}\\.[[:digit:]]{1,3}\\.[[:digit:]]{1,3})";
  const char *str_regex_port = "([[:digit:]]{2,5})";
  err_addr = regcomp (&preg_addr, str_regex_addr, REG_EXTENDED);
  err_port = regcomp (&preg_port, str_regex_port, REG_EXTENDED);
  pid_t pid;
  FILE *fo;
  if ((err_addr == 0) && (err_port == 0)) {
    int match;
    size_t nmatch_addr = 0, nmatch_port = 0;
    regmatch_t *pmatch_addr = NULL, *pmatch_port = NULL;

    nmatch_addr = preg_addr.re_nsub;
    nmatch_port = preg_port.re_nsub;
    pmatch_addr = malloc (sizeof (*pmatch_addr) * nmatch_addr);
    pmatch_port = malloc (sizeof (*pmatch_port) * nmatch_port);
    if (pmatch_addr && pmatch_port) {
      struct proxy *p;
      int start=0, end=0;
      size_t size;
      p=malloc(sizeof(struct proxy));
      fo=fopen(fileout,"w+");
      while ((match=regexec (&preg_addr, text_to_parse, nmatch_addr, pmatch_addr, 0)) == 0) {
        start = pmatch_addr->rm_so;
        end = pmatch_addr->rm_eo;
        size = end - start;
        memset(p->ip,0,sizeof(p->ip));
        strncpy (p->ip, &text_to_parse[start], size);
        p->ip[size] = '\0';

        text_to_parse+=end;

        if((match=regexec (&preg_port, text_to_parse,nmatch_port,pmatch_port,0)) == 0) {
          start = pmatch_port->rm_so;
          end = pmatch_port->rm_eo;
          size = end - start;
          memset(p->port,0,sizeof(p->port));
          strncpy (p->port, &text_to_parse[start], size);
          p->port[size] = '\0';
          if (atoi(p->port)>=80 && atoi(p->port)<=65535) {
            if((pid=fork())==-1) {
              printf("[-] fork error\n");
              exit(1);
            } else if (pid==0) {
              if(check_ssl(p)==0)
                fprintf(fo,"%s:%s\n", p->ip, p->port);
              exit (0);
            }
          }
          text_to_parse+=end;
        } else return(EXIT_SUCCESS);
      }
      free(p);
    } else {
      fprintf (stderr, "[-] Not enough memory\n");
      exit (EXIT_FAILURE);
    }
  }
  sleep(CHECK_TIMEOUT);
  fclose(fo);

  return (EXIT_SUCCESS);
}
